library(tidyverse)
library(rstan)
library(bayesplot)

set.seed(2020)

#setwd('/Users/walkerharrison/Downloads/')
setwd('P:/sharks/')

sharks <- read_csv('masterDataFrame.csv')

# create session index
sharks <- sharks %>%
  mutate(session_char = as.factor(paste0(dayCount, dawnDusk, Location))) %>%
  mutate(session = as.numeric(session_char))

# continuous and categorical predictors
pred_cont <- c('WSPD', 'PRES', 'WTMP', 'tideHeight', 'relativeStart')
pred_cat <- c('dawnDusk', 'Location')
pred_all <- c(pred_cont, pred_cat)
n_pred <- length(pred_all)


#################################################
###############  Model Selection  ###############
#################################################

# turn this on to perform model selection
# takes 5-10 hours across 12 cores
select_model <- FALSE

if(select_model){
  
  library(parallel)
  library(doSNOW)
  
  # indices for every possible subset
  subsets <- sapply(1:n_pred, 
                    function(n) combn(1:n_pred, n, simplify = FALSE)) %>%
    unlist(recursive = FALSE)
  
  
  # models dataframe for record keeping
  models <- data.frame(id = 1:length(subsets)) %>%
    mutate(subset_idx = subsets) %>%
    rowwise() %>%
    mutate(predictors = list(pred_all[subset_idx])) %>%
    ungroup() %>%
    mutate(elpd = NA)
  
  # set up parallel processing
  cores <- parallel::detectCores()
  cl <- makeCluster(cores-1)
  registerDoSNOW(cl)
  
  # progress bar 
  pb <- txtProgressBar(max = max(sharks$session), style = 3)
  progress <- function(n) setTxtProgressBar(pb, n)
  opts <- list(progress = progress)
  
  # loop over each predictor subset
  for(i in nrow(models):1){
    print(paste0("Testing Model ", i, " of ", nrow(models)))
    
    # model matrix using predictor indices
    pred_formula <- as.formula(paste0("~ ", paste0(models$predictors[[i]], collapse = " + ")))
    X <- model.matrix(pred_formula, data = sharks)
    
    # scale continuous predictors
    cont_idx <- which(colnames(X) %in% pred_cont)
    X[, cont_idx] <- scale(X[, cont_idx])
    
    # parallel loop over left out groups
    lls <- foreach(j = 1:max(sharks$session),
                   .options.snow = opts,
                   .combine = 'c',
                   .packages = c('rstan')) %dopar%{
                     
                     # catch fitting fails (likely with so many Stan runs)
                     tryCatch({
                       fit <- stan(file = "negbin_final2.stan",
                                   
                                   data = list(
                                     N = nrow(sharks),
                                     N_sessions = max(sharks$session),
                                     session = sharks$session,
                                     count = sharks$count,
                                     area = sharks$areaCovered*100,
                                     K = ncol(X),
                                     X = X,
                                     LOGO = j
                                   ),
                                   
                                   iter = 1000,
                                   warmup = 500,
                                   chains = 1,
                                   refresh = 0,
                                   seed = 2020)
                       
                       # return mean posterior log predictive density
                       mean(extract(fit)$log_lik)
                       
                     },
                     error = function(e){
                       return(NA)
                     })
                   }
    
    # record log predictive density if it didn't fail
    if(sum(is.na(lls)) == 0){
      models$elpd[i] <- sum(lls)
    }
    
    # save progress
    save(models, file = "models.Rdata")
  }
  
  # show best models by ELPD
  models %>%
    arrange(desc(elpd)) %>%
    head(5)
}


#################################################
#################  Final Model  #################
#################################################

# model matrix using best predictor subset
X <- model.matrix(~WSPD + WTMP + relativeStart + Location, data = sharks)

# scale continuous predictors
cont_idx <- which(colnames(X) %in% pred_cont)
X[, cont_idx] <- scale(X[, cont_idx])

# fit model
fit <- stan(file = "negbin_final2.stan",
            
            data = list(
              N = nrow(sharks),
              N_sessions = max(sharks$session),
              session = sharks$session,
              count = sharks$count,
              area = sharks$areaCovered*100,
              K = ncol(X),
              X = X,
              LOGO = max(sharks$session) + 1
            ),
            
            iter = 2000,
            warmup = 1000,
            chains = 4,
            refresh = 0,
            seed = 2020)

# show summary stats of posterior draws of parameters
summary(fit, pars = c("beta", "sigma", "phi"))$summary

# make pairs plots to diagnose any sampling issues
pairs(fit, pars = c("beta", "sigma", "phi"))

posterior <- extract(fit)

# range and effect of random effects
re <- round(apply(posterior$u, 2, mean), 3)
range(re)

median_re <- 12
re[median_re]

preds <- apply(posterior$count_rep, 2, mean)
preds[sharks$session == median_re]

exp(log(preds[sharks$session == median_re]) - re[median_re])

# show unit (or 10-minute) change in predictors for baseline abundance of 8 sharks
exp(log(8) + apply(posterior$beta, 2, mean)[2]/sd(sharks$WSPD))
exp(log(8) + apply(posterior$beta, 2, mean)[3]/sd(sharks$WTMP))
exp(log(8) + 10*apply(posterior$beta, 2, mean)[4]/sd(sharks$relativeStart))
exp(log(8) + apply(posterior$beta, 2, mean)[5])


# loop over left out groups for out-of-sample predictions
fits <- lapply(1:max(sharks$session), function(j){
  
  stan(file = "negbin_final2.stan",
       
       data = list(
         N = nrow(sharks),
         N_sessions = max(sharks$session),
         session = sharks$session,
         count = sharks$count,
         area = sharks$areaCovered*100,
         K = ncol(X),
         X = X,
         LOGO = j
       ),
       
       iter = 2000,
       warmup = 1000,
       chains = 4,
       refresh = 0,
       seed = 2020)
})

# extract out-of-sample predictions for each model
cv_preds <- unlist(lapply(1:max(sharks$session), function(i){
  apply(extract(fits[[i]])$count_rep, 2, mean)[sharks$session == i]
}))

# loss functions for predictions
sharks %>% 
  arrange(session, Transect) %>% 
  mutate(pred_count = cv_preds) %>%
  summarize(RMSE = sqrt(mean((count-pred_count)^2)),
            MAE = median(abs(count-pred_count)))

theme_set(theme_bw())

# fixed effects posterior distributions
g1 <- mcmc_areas(extract(fit, permuted = FALSE), pars = c("beta[1]",
                                                          "beta[2]",
                                                          "beta[3]",
                                                          "beta[4]",
                                                          "beta[5]"),
                 prob = 0.95,
                 point_est = "mean") +
  labs(
    title = "Posterior distributions",
    subtitle = "with means and 95% intervals"
  )

# out of sample predictions vs actual
g2 <- sharks %>% 
  arrange(session, Transect) %>% 
  mutate(pred_count = cv_preds) %>%
  ggplot(aes(pred_count, count)) +
  geom_point() +
  geom_abline() +
  theme_bw() +
  labs(x = "Predicted Count (out of sample)",
       y = "Actual Count")

g1

g2
